@@ -6,9 +6,9 @@ import json
 import itertools as it
 import logging
 import scipy.sparse
-#from tqdm import tqdm
+from tqdm import tqdm
 import time
-#import importlib.resources as pkg_resources
+import importlib.resources as pkg_resources
 
 # sklearn
 import sklearn
